from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.ecosystem_data.arch import AlpmDBEntry
from labels.model.file import Location
from labels.model.package import Language, Package, PackageType
from labels.model.release import Release
from labels.parsers.cataloger.utils import (
    get_enriched_location,
    log_malformed_package_warning,
    purl_qualifiers,
)


def new_arch_package(
    entry: AlpmDBEntry, release: Release | None, db_location: Location
) -> Package | None:
    name = entry.package
    version = entry.version

    if not name or not version:
        return None

    new_location = get_enriched_location(db_location)
    p_url = _get_arch_package_url(entry, release)

    try:
        return Package(
            name=name,
            version=version,
            locations=[new_location],
            type=PackageType.AlpmPkg,
            ecosystem_data=entry,
            p_url=p_url,
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        log_malformed_package_warning(new_location, ex)
        return None


def _get_arch_package_url(entry: AlpmDBEntry, distro: Release | None = None) -> str:
    qualifiers: dict[str, str | None] = {"arch": entry.architecture}
    if entry.base_package:
        qualifiers["upstream"] = entry.base_package

    return PackageURL(  # type: ignore[misc]
        type="alpm",
        name=entry.package,
        version=entry.version,
        qualifiers=purl_qualifiers(qualifiers, distro),
    ).to_string()



import hashlib
import json
from enum import Enum, StrEnum
from typing import cast

from pydantic import BaseModel

from labels.model.advisories import Advisory
from labels.model.constraints import NonEmptyStr, TrimmedNonEmptyStr
from labels.model.ecosystem_data.aliases import AcceptedEcosystemData
from labels.model.file import Location
from labels.model.metadata import HealthMetadata


class Platform(StrEnum):
    ALPM = "ALPM"
    APK = "ALPINE"
    BINARY = "BINARY"
    CABAL = "CABAL"
    COMPOSER = "COMPOSER"
    CRAN = "CRAN"
    DEBIAN = "DEBIAN"
    GEM = "GEM"
    GO = "GO"
    GRAALVM_NATIVE_IMAGE = "GRAALVM_NATIVE_IMAGE"
    JENKINS_PLUGIN = "JENKINS_PLUGIN"
    LINUX_KERNEL = "LINUX_KERNEL"
    LINUX_KERNEL_MODULE = "LINUX_KERNEL_MODULE"
    MAVEN = "MAVEN"
    MSRC_KB = "MSRC_KB"
    NIXPKGS = "NIXPKGS"
    NPM = "NPM"
    NUGET = "NUGET"
    PIP = "PIP"
    PORTAGE = "PORTAGE"
    PUB = "PUB"
    RPM = "RPM"
    SWIFT = "SWIFT"
    UNKNOWN = "UNKNOWN"
    WORDPRESS_PLUGIN = "WORDPRESS_PLUGIN"


class Ecosystem(StrEnum):
    ALPM = "alpm"
    ALPINE = "alpine"
    BINARY = "binary"
    CABAL = "cabal"
    CRAN = "cran"
    DEBIAN = "debian"
    GO = "go"
    GRAALVM_NATIVE_IMAGE = "graalvm_native_image"
    JENKINS_PLUGIN = "jenkins_plugin"
    LINUX_KERNEL = "linux_kernel"
    LINUX_KERNEL_MODULE = "linux_kernel_module"
    MAVEN = "maven"
    MSRC_KB = "msrc_kb"
    NIXPKGS = "nixpkgs"
    NPM = "npm"
    NUGET = "nuget"
    PACKAGIST = "packagist"
    PORTAGE = "portage"
    PUB = "pub"
    PYPI = "pypi"
    RPM = "rpm"
    RUBYGEMS = "rubygems"
    SWIFTURL = "swifturl"
    WORDPRESS_PLUGIN = "wordpress_plugin"


class Language(StrEnum):
    UNKNOWN_LANGUAGE = "unknown_language"
    DART = "dart"
    DOTNET = "dotnet"
    GO = "go"
    HASKELL = "haskell"
    JAVA = "java"
    JAVASCRIPT = "javascript"
    PHP = "php"
    PYTHON = "python"
    R = "R"
    RUBY = "ruby"
    SWIFT = "swift"


class PackageType(Enum):
    UnknownPkg = "UnknownPackage"
    AlpmPkg = "alpm"
    ApkPkg = "apk"
    BinaryPkg = "binary"
    CocoapodsPkg = "pod"
    DartPubPkg = "dart-pub"
    DebPkg = "deb"
    DotnetPkg = "dotnet"
    GemPkg = "gem"
    GemspecPkg = "gemspec"
    GoModulePkg = "go-module"
    GraalVMNativeImagePkg = "graalvm-native-image"
    HackagePkg = "hackage"
    JavaPkg = "java-archive"
    JenkinsPluginPkg = "jenkins-plugin"
    KbPkg = "msrc-kb"
    LinuxKernelPkg = "linux-kernel"
    LinuxKernelModulePkg = "linux-kernel-module"
    NixPkg = "nix"
    NpmPkg = "npm"
    PhpComposerPkg = "php-composer"
    PhpPeclPkg = "php-pecl-pkg"
    PhpPearPkg = "php-pear"
    PortagePkg = "portage"
    PythonPkg = "python"
    Rpkg = "R-package"
    RpmPkg = "rpm"
    SwiftPkg = "swift"
    WordpressPluginPkg = "wordpress-plugin"

    def get_platform_value(self) -> str | None:
        package_type_to_platform = {
            PackageType.CocoapodsPkg: Platform.SWIFT.value,
            PackageType.DartPubPkg: Platform.PUB.value,
            PackageType.DotnetPkg: Platform.NUGET.value,
            PackageType.GemPkg: Platform.GEM.value,
            PackageType.GemspecPkg: Platform.GEM.value,
            PackageType.GoModulePkg: Platform.GO.value,
            PackageType.HackagePkg: Platform.CABAL.value,
            PackageType.JavaPkg: Platform.MAVEN.value,
            PackageType.NpmPkg: Platform.NPM.value,
            PackageType.PhpComposerPkg: Platform.COMPOSER.value,
            PackageType.PhpPeclPkg: Platform.COMPOSER.value,
            PackageType.PythonPkg: Platform.PIP.value,
            PackageType.Rpkg: Platform.CRAN.value,
            PackageType.SwiftPkg: Platform.SWIFT.value,
            PackageType.AlpmPkg: Platform.ALPM.value,
            PackageType.ApkPkg: Platform.APK.value,
            PackageType.BinaryPkg: Platform.BINARY.value,
            PackageType.DebPkg: Platform.DEBIAN.value,
            PackageType.GraalVMNativeImagePkg: Platform.GRAALVM_NATIVE_IMAGE.value,
            PackageType.JenkinsPluginPkg: Platform.JENKINS_PLUGIN.value,
            PackageType.KbPkg: Platform.MSRC_KB.value,
            PackageType.LinuxKernelPkg: Platform.LINUX_KERNEL.value,
            PackageType.LinuxKernelModulePkg: Platform.LINUX_KERNEL_MODULE.value,
            PackageType.NixPkg: Platform.NIXPKGS.value,
            PackageType.PortagePkg: Platform.PORTAGE.value,
            PackageType.RpmPkg: Platform.RPM.value,
            PackageType.WordpressPluginPkg: Platform.WORDPRESS_PLUGIN.value,
            PackageType.UnknownPkg: None,
        }
        return package_type_to_platform.get(self, Platform.UNKNOWN.value)

    def get_ecosystem_value(self) -> str | None:
        package_type_to_ecosystem: dict[PackageType, str | None] = {
            PackageType.AlpmPkg: Ecosystem.ALPM,
            PackageType.ApkPkg: Ecosystem.ALPINE,
            PackageType.BinaryPkg: Ecosystem.BINARY,
            PackageType.CocoapodsPkg: Ecosystem.SWIFTURL,
            PackageType.DartPubPkg: Ecosystem.PUB,
            PackageType.DebPkg: Ecosystem.DEBIAN,
            PackageType.DotnetPkg: Ecosystem.NUGET,
            PackageType.GemPkg: Ecosystem.RUBYGEMS,
            PackageType.GemspecPkg: Ecosystem.RUBYGEMS,
            PackageType.GoModulePkg: Ecosystem.GO,
            PackageType.GraalVMNativeImagePkg: Ecosystem.GRAALVM_NATIVE_IMAGE,
            PackageType.HackagePkg: Ecosystem.CABAL,
            PackageType.JavaPkg: Ecosystem.MAVEN,
            PackageType.JenkinsPluginPkg: Ecosystem.JENKINS_PLUGIN,
            PackageType.KbPkg: Ecosystem.MSRC_KB,
            PackageType.LinuxKernelPkg: Ecosystem.LINUX_KERNEL,
            PackageType.LinuxKernelModulePkg: Ecosystem.LINUX_KERNEL_MODULE,
            PackageType.NixPkg: Ecosystem.NIXPKGS,
            PackageType.NpmPkg: Ecosystem.NPM,
            PackageType.PhpComposerPkg: Ecosystem.PACKAGIST,
            PackageType.PhpPeclPkg: Ecosystem.PACKAGIST,
            PackageType.PhpPearPkg: Ecosystem.PACKAGIST,
            PackageType.PortagePkg: Ecosystem.PORTAGE,
            PackageType.PythonPkg: Ecosystem.PYPI,
            PackageType.Rpkg: Ecosystem.CRAN,
            PackageType.RpmPkg: Ecosystem.RPM,
            PackageType.SwiftPkg: Ecosystem.SWIFTURL,
            PackageType.WordpressPluginPkg: Ecosystem.WORDPRESS_PLUGIN,
            PackageType.UnknownPkg: None,
        }
        return package_type_to_ecosystem.get(self)


class Package(BaseModel):
    name: NonEmptyStr
    version: TrimmedNonEmptyStr
    language: Language
    locations: list[Location]
    type: PackageType
    advisories: list[Advisory] | None = None
    found_by: NonEmptyStr | None = None
    health_metadata: HealthMetadata | None = None
    ecosystem_data: AcceptedEcosystemData | None = None
    p_url: NonEmptyStr
    safe_versions: list[str] | None = None
    syft_id: NonEmptyStr | None = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Package):
            return False

        self_dump = cast("dict[str, object]", self.model_dump())
        other_dump = cast("dict[str, object]", other.model_dump())

        return self_dump == other_dump

    def __hash__(self) -> int:
        return hash(self.id_)

    @property
    def id_(self) -> str:
        return self.id_by_hash()

    def id_by_hash(self) -> str:
        try:
            obj_data = {
                "name": self.name,
                "version": self.version,
                "language": self.language.value,
                "type": self.type.value,
                "p_url": self.p_url,
            }
            obj_str = json.dumps(obj_data, sort_keys=True)
            return hashlib.sha256(obj_str.encode()).hexdigest()
        except Exception as exc:
            error_message = "Package data invalid for ID generation: " + json.dumps(
                obj_data, sort_keys=True
            )
            raise ValueError(error_message) from exc



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.parsers.cataloger.utils import log_malformed_package_warning


def new_npm_package_from_lock(
    location: Location,
    name: str | None,
    value: IndexedDict[str, ParsedValue],
    lockfile_version: int,
) -> Package | None:
    version: str = str(value.get("version", ""))

    if not name or not version:
        return None

    if lockfile_version == 1:
        alias_prefix_package_lock = "npm:"
        if version.startswith(alias_prefix_package_lock):
            name, version = version.removeprefix(alias_prefix_package_lock).rsplit(
                "@",
                1,
            )

    p_url = _get_npm_package_url(name, version)

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.JAVASCRIPT,
            type=PackageType.NpmPkg,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def new_simple_npm_package(
    location: Location, name: str | None, version: str | None
) -> Package | None:
    if not name or not version:
        return None

    p_url = _get_npm_package_url(name, version)

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.JAVASCRIPT,
            type=PackageType.NpmPkg,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(location, ex)
        return None


def _get_npm_package_url(name: str, version: str) -> str:
    namespace = ""
    fields = name.split("/", 2)
    if len(fields) > 1:
        namespace = fields[0]
        name = fields[1]

    if not name:
        return ""

    return PackageURL(type="npm", namespace=namespace, name=name, version=version).to_string()  # type: ignore[misc]



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.ecosystem_data.redhat import RpmDBEntry
from labels.model.file import Location
from labels.model.package import Language, Package, PackageType
from labels.model.release import Environment
from labels.parsers.cataloger.redhat.rpmdb.domain.package import PackageInfo
from labels.parsers.cataloger.utils import get_enriched_location, log_malformed_package_warning


def new_redhat_package(
    *, entry: PackageInfo, location: Location, env: Environment
) -> Package | None:
    name = entry.name
    version = entry.version

    if not name or not version:
        return None

    ecosystem_data = RpmDBEntry(
        name=name,
        version=version,
        epoch=entry.epoch,
        arch=entry.arch,
        release=entry.release,
        source_rpm=entry.source_rpm,
    )

    new_location = get_enriched_location(location)
    p_url = _package_url(entry, env)

    el_version = _to_el_version(entry.epoch, version, entry.release)

    try:
        return Package(
            name=name,
            version=el_version,
            locations=[new_location],
            language=Language.UNKNOWN_LANGUAGE,
            type=PackageType.RpmPkg,
            ecosystem_data=ecosystem_data,
            p_url=p_url,
        )
    except ValidationError as ex:
        log_malformed_package_warning(new_location, ex)
        return None


def _package_url(entry: PackageInfo, env: Environment) -> str:
    namespace = ""
    if env.linux_release:
        namespace = env.linux_release.id_
    qualifiers: dict[str, str] = {}
    if entry.arch:
        qualifiers["arch"] = entry.arch
    if entry.epoch:
        qualifiers["epoch"] = str(entry.epoch)
    if entry.source_rpm:
        qualifiers["upstream"] = entry.source_rpm

    return PackageURL(
        type="rpm",
        namespace=namespace,
        name=entry.name,
        version=f"{entry.version}-{entry.release}",
        qualifiers=qualifiers,
        subpath="",
    ).to_string()


def _to_el_version(epoch: int | None, version: str, release: str) -> str:
    if epoch:
        return f"{epoch}:{version}-{release}"
    return f"{version}-{release}"



from labels.model.ecosystem_data.arch import AlpmDBEntry
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment, Release
from labels.model.resolver import Resolver
from labels.parsers.cataloger.arch.package_builder import new_arch_package

ALLOWED_STRING_KEYS: set[str] = {"name", "base", "version", "arch"}


def parse_alpm_db(
    _resolver: Resolver,
    environment: Environment,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    alpm_db_entry = _collect_alpm_db_entry(reader.read_closer.read())
    if not alpm_db_entry:
        return [], []

    packages = _collect_packages(alpm_db_entry, environment.linux_release, reader.location)

    return packages, []


def _collect_packages(
    alpm_db_entry: AlpmDBEntry, linux_release: Release | None, location: Location
) -> list[Package]:
    packages: list[Package] = []

    package = new_arch_package(alpm_db_entry, linux_release, location)
    if package:
        packages.append(package)

    return packages


def _collect_alpm_db_entry(content: str) -> AlpmDBEntry | None:
    pkg_fields: dict[str, str] = {}
    lines = content.split("\n\n")

    for line in lines:
        if not line.strip():
            break  # End of block or file
        pkg_fields.update(_parse_key_value_pair(line))

    if pkg_fields:
        return _parse_raw_entry(pkg_fields)

    return None


def _parse_raw_entry(pkg_fields: dict[str, str]) -> AlpmDBEntry:
    return AlpmDBEntry(
        package=pkg_fields.get("name", ""),
        base_package=pkg_fields.get("base", ""),
        version=pkg_fields.get("version", ""),
        architecture=pkg_fields.get("arch", ""),
    )


def _parse_key_value_pair(line: str) -> dict[str, str]:
    try:
        key, value = line.split("\n", 1)
    except ValueError:
        return {}

    key = key.replace("%", "").lower()
    value = value.strip()

    if key not in ALLOWED_STRING_KEYS:
        return {}

    return {key: value}



from more_itertools import flatten

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package, PackageType
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.dotnet.utils import build_relationships_from_json_dependencies
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter
from labels.utils.strings import normalize_name


def parse_dotnet_package_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    targets = _collect_targets(file_content)
    if not targets:
        return [], []

    packages = _collect_packages(targets, reader.location)
    relationships = _collect_relationships(targets, packages)

    return packages, relationships


def _collect_targets(
    file_content: IndexedDict[str, ParsedValue],
) -> list[tuple[str, ParsedValue]] | None:
    targets: ParsedValue = file_content.get("dependencies")
    if not isinstance(targets, IndexedDict):
        return None

    return list(
        flatten(target.items() for target in targets.values() if isinstance(target, IndexedDict))
    )


def _collect_packages(targets: list[tuple[str, ParsedValue]], location: Location) -> list[Package]:
    packages: list[Package] = []

    for package_name, package_value in targets:
        if not isinstance(package_value, IndexedDict):
            continue

        version = package_value.get("resolved")
        if not isinstance(version, str):
            continue

        is_transitive = package_value.get("type", "") == "Transitive"
        new_location = get_enriched_location(
            location, line=package_value.position.start.line, is_transitive=is_transitive
        )

        new_package = new_dotnet_package(package_name, version, new_location)
        if new_package:
            packages.append(new_package)

    return packages


def _collect_relationships(
    targets: list[tuple[str, ParsedValue]],
    packages: list[Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    packages_by_key = {(pkg.name, pkg.version): pkg for pkg in packages}

    for package_key, package_value in targets:
        relationships.extend(
            _build_relationships_for_package(package_key, package_value, packages_by_key)
        )

    return relationships


def _build_relationships_for_package(
    package_key: str,
    package_value: ParsedValue,
    packages_by_key: dict[tuple[str, str], Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not isinstance(package_value, IndexedDict):
        return relationships

    normalized_name = normalize_name(package_key, PackageType.DotnetPkg)
    version = package_value.get("resolved")
    if not isinstance(version, str):
        return relationships

    current_package = packages_by_key.get((normalized_name, version))
    if not isinstance(current_package, Package):
        return relationships

    dependencies = package_value.get("dependencies")
    if not isinstance(dependencies, IndexedDict):
        return relationships

    return build_relationships_from_json_dependencies(
        dependencies, current_package, packages_by_key
    )



import logging
import struct

LOGGER = logging.getLogger(__name__)


def htonl(val: int) -> int:
    try:
        # Convert from little-endian (host byte order)
        # to big-endian (network byte order)
        return int(struct.unpack(">i", struct.pack("<i", val))[0])
    except struct.error:
        LOGGER.exception("Failed to convert integer: %s")
        return 0


def htonlu(val: int) -> int:
    try:
        # Convert from little-endian (host byte order)
        # to big-endian (network byte order)
        return int(struct.unpack(">I", struct.pack("<I", val))[0])
    except struct.error:
        LOGGER.exception("Failed to convert unsigned integer: %s")
        return 0



from typing import cast

from bs4 import BeautifulSoup, Tag

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.utils import get_enriched_location

PACKAGE_TAG = "package"


def parse_dotnet_pkgs_config(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = BeautifulSoup(reader.read_closer.read(), features="html.parser")

    packages = _collect_packages(file_content, reader.location)

    return packages, []


def _collect_packages(file_content: BeautifulSoup, location: Location) -> list[Package]:
    packages: list[Package] = []

    packages_tags: list[Tag] = file_content.find_all(PACKAGE_TAG, recursive=True)
    for raw_package in packages_tags:
        name = cast("str | None", raw_package.get("id"))
        version = cast("str | None", raw_package.get("version"))

        new_location = get_enriched_location(
            location, line=raw_package.sourceline, is_transitive=False
        )

        new_package = new_dotnet_package(name, version, new_location)
        if new_package:
            packages.append(new_package)

    return packages



import re
from typing import cast

from bs4 import BeautifulSoup, Tag

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.utils import get_enriched_location

PACKAGE = re.compile(r".+\\packages\\(?P<package_info>[^\s\\]*)\\.+")
DEP_INFO = re.compile(r"(?P<package_name>.*?)\.(?P<version>\d+[^\s]*)$")


PACKAGE_REFERENCE_ELEMENT = "packagereference"
PRIVATE_ASSETS_ELEMENT = "privateassets"
REFERENCE_ELEMENT = "reference"
HINT_PATH_ELEMENT = "hintpath"


def parse_csproj(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = BeautifulSoup(reader.read_closer.read(), features="html.parser")

    packages = _collect_packages(file_content, reader)

    return packages, []


def _collect_packages(root: BeautifulSoup, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []

    package_references: list[Tag] = root.find_all(PACKAGE_REFERENCE_ELEMENT, recursive=True)
    for raw_package in package_references:
        package_name = cast("str | None", raw_package.get("include"))
        version = cast("str | None", raw_package.get("version"))

        is_dev = _is_dev_dependency(raw_package)
        new_location = get_enriched_location(
            reader.location,
            line=raw_package.sourceline,
            is_transitive=False,
            is_dev=is_dev,
        )

        package = new_dotnet_package(package_name, version, new_location)
        if package:
            packages.append(package)

    packages.extend(_collect_package_from_reference_dependencies(root, reader))

    return packages


def _is_dev_dependency(pkg: Tag) -> bool:
    checking_attr = "all"

    element = pkg.find(PRIVATE_ASSETS_ELEMENT)
    if isinstance(element, Tag):
        return element.get_text(strip=True).lower() == checking_attr

    attr = pkg.get(PRIVATE_ASSETS_ELEMENT)
    if isinstance(attr, str):
        return attr.strip().lower() == checking_attr

    return False


def _collect_package_from_reference_dependencies(
    raw_package: Tag, reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []
    references: list[Tag] = raw_package.find_all(REFERENCE_ELEMENT, recursive=True)

    for reference in references:
        package = _extract_package_from_reference(reference, reader)
        if package:
            packages.append(package)

    return packages


def _extract_package_from_reference(reference: Tag, reader: LocationReadCloser) -> Package | None:
    package = _extract_from_hint_path(reference, reader)
    if package:
        return package

    return _extract_from_include_attribute(reference, reader)


def _extract_from_hint_path(reference: Tag, reader: LocationReadCloser) -> Package | None:
    dll_path = reference.find(HINT_PATH_ELEMENT)
    if not isinstance(dll_path, Tag):
        return None

    dll_package = PACKAGE.match(dll_path.text)
    if not dll_package:
        return None

    package_info = DEP_INFO.match(dll_package.group("package_info"))
    if not package_info:
        return None

    new_location = get_enriched_location(
        reader.location, line=dll_path.sourceline, is_transitive=False
    )

    name: str = package_info.group("package_name").lower()
    version: str = package_info.group("version")

    return new_dotnet_package(name, version, new_location)


def _extract_from_include_attribute(raw_package: Tag, reader: LocationReadCloser) -> Package | None:
    include = raw_package.get("include")
    if not isinstance(include, str):
        return None

    include_info = include.replace(" ", "").split(",")

    package_name = include_info[0].strip()
    version = _get_version(include_info)

    if not package_name or not version:
        return None

    new_location = get_enriched_location(
        reader.location, line=raw_package.sourceline, is_transitive=False
    )

    return new_dotnet_package(package_name, version, new_location)


def _get_version(include_info: list[str]) -> str | None:
    return next(
        (
            package_info.lstrip("Version=")
            for package_info in include_info
            if package_info.startswith("Version=")
        ),
        None,
    )



from pathlib import Path
from typing import NamedTuple, TextIO, cast

from bs4 import BeautifulSoup, Tag

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProperties
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.java.utils.model import PomContext
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.java.utils.parsing_utils import get_direct_child_text
from labels.parsers.cataloger.java.utils.version_resolver import resolve_version
from labels.parsers.cataloger.utils import get_enriched_location


class PomEvaluationContext(NamedTuple):
    resolver: Resolver
    parent_info: dict[str, str]
    current_project: Tag
    current_pom_path: str


def parse_pom_xml(
    resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    pom_tree = _safe_parse_pom_xml(reader)
    if not pom_tree:
        return [], []

    pom_context = _build_pom_context(pom_tree, reader, resolver)
    if not pom_context:
        return [], []

    packages = _collect_packages(pom_context, reader.location)

    return packages, []


def _safe_parse_pom_xml(reader: LocationReadCloser) -> BeautifulSoup | None:
    try:
        return BeautifulSoup(reader.read_closer, features="html.parser")
    except (AssertionError, UnicodeError):
        return None


def _build_pom_context(
    pom_tree: BeautifulSoup,
    reader: LocationReadCloser,
    resolver: Resolver | None,
) -> PomContext | None:
    valid_project_and_dependencies = _get_valid_project_and_dependencies(pom_tree)
    if not valid_project_and_dependencies:
        return None

    project, dependencies = valid_project_and_dependencies

    parent_info = _get_parent_info(project)

    if resolver and parent_info:
        evaluation_context = PomEvaluationContext(
            resolver=resolver,
            parent_info=parent_info,
            current_project=project,
            current_pom_path=str(reader.location.access_path),
        )
        parent_version_properties, manage_deps = _evaluate_pom_files_in_project(evaluation_context)
    else:
        parent_version_properties = _get_properties(pom_tree)
        manage_deps = _get_deps_management(pom_tree)

    return PomContext(
        project=project,
        dependencies=dependencies,
        parent_info=parent_info,
        parent_version_properties=parent_version_properties,
        manage_deps=manage_deps,
    )


def _get_valid_project_and_dependencies(pom_tree: BeautifulSoup) -> tuple[Tag, Tag] | None:
    project = getattr(pom_tree, "project", None)
    if not project:
        return None

    if str(project.get("xmlns")) != "http://maven.apache.org/POM/4.0.0":
        return None

    dependencies = project.find("dependencies", recursive=False)
    if not isinstance(dependencies, Tag):
        return None

    return project, dependencies


def _get_parent_info(project: Tag) -> dict[str, str] | None:
    parent = project.find_next("parent")
    if not isinstance(parent, Tag):
        return None

    group_node = get_direct_child_text(parent, "groupid")
    artifact_node = get_direct_child_text(parent, "artifactid")
    version_node = get_direct_child_text(parent, "version")

    if not (group_node and artifact_node and version_node):
        return None

    return {
        "group": group_node,
        "artifact": artifact_node,
        "version": version_node,
    }


def _evaluate_pom_files_in_project(
    context: PomEvaluationContext,
) -> tuple[dict[str, str], dict[str, str]]:
    parent_pom = _find_parent_pom(context)
    properties_vars = _get_properties(parent_pom) if parent_pom else {}
    manage_deps = _get_deps_management(parent_pom) if parent_pom else {}

    manage_deps.update(_get_deps_management(context.current_project))

    return properties_vars, manage_deps


def _find_parent_pom(context: PomEvaluationContext) -> Tag | None:
    for loc in context.resolver.files_by_glob("**/pom.xml", "pom.xml"):
        content = context.resolver.file_contents_by_location(loc)
        if not content:
            continue

        pom_project = _get_pom_project(content)
        if not pom_project:
            continue

        if _is_parent_pom(pom_project, context.parent_info) and _is_module_parent(
            pom_project, context.current_pom_path
        ):
            return pom_project

    return None


def _get_pom_project(content: TextIO) -> Tag | None:
    file_content = BeautifulSoup(content, features="html.parser")
    pom_project = getattr(file_content, "project", None)

    if (
        pom_project
        and isinstance(pom_project, Tag)
        and str(pom_project.get("xmlns")) == "http://maven.apache.org/POM/4.0.0"
    ):
        return cast("Tag", pom_project)

    return None


def _is_parent_pom(root_pom_project: Tag, parent_info: dict[str, str]) -> bool:
    group = root_pom_project.find("groupid", recursive=False)
    artifact = root_pom_project.find("artifactid", recursive=False)
    version = root_pom_project.find("version", recursive=False)

    if not (group and artifact and version):
        return False

    return (
        group.get_text() == parent_info["group"]
        and artifact.get_text() == parent_info["artifact"]
        and version.get_text() == parent_info["version"]
    )


def _is_module_parent(parent_pom: Tag, pom_module: str) -> bool:
    base_module_name = Path(pom_module).parent.name

    for modules_section in parent_pom.find_all("modules"):
        for module_tag in modules_section.find_all("module"):
            if module_tag.get_text() == base_module_name:
                return True

    return False


def _get_properties(project_pom: Tag) -> dict[str, str]:
    properties_dict: dict[str, str] = {}

    for properties_section in project_pom.find_all("properties", limit=2):
        for prop in properties_section.children:
            if isinstance(prop, Tag) and prop.name:
                properties_dict[prop.name.lower()] = prop.get_text()

    return properties_dict


def _get_deps_management(pom_tree: Tag) -> dict[str, str]:
    deps_info: dict[str, str] = {}

    for manage in pom_tree.find_all("dependencymanagement"):
        for dependency in manage.find_all("dependency", recursive=True):
            if not (dependency.groupid and dependency.artifactid and dependency.version):
                continue

            group = dependency.groupid.get_text()
            artifact = dependency.artifactid.get_text()
            version = dependency.version.get_text()
            key = f"{group}:{artifact}"
            deps_info[key] = version

    return deps_info


def _collect_packages(pom_context: PomContext, location: Location) -> list[Package]:
    packages: list[Package] = []

    for dependency in pom_context.dependencies.find_all("dependency"):
        package_spec = _build_package_spec(pom_context, dependency, location)
        if not package_spec:
            continue

        package = new_java_package(package_spec)
        if package:  # pragma: no branch
            packages.append(package)

    return packages


def _build_package_spec(
    pom_context: PomContext, dependency: Tag, location: Location
) -> JavaPackageSpec | None:
    name = get_direct_child_text(dependency, "artifactid")
    if not name:
        return None

    group_id = get_direct_child_text(dependency, "groupid")
    full_name = f"{group_id}:{name}" if group_id else name

    pom_properties = JavaPomProperties(
        group_id=group_id,
        artifact_id=name,
        version=get_direct_child_text(dependency, "version"),
    )

    version = resolve_version(pom_context, pom_properties, full_name)
    if not _is_valid_version(version):
        return None

    scope = get_direct_child_text(dependency, "scope")
    is_dev = scope in ["test", "provided"]

    new_location = get_enriched_location(
        location,
        line=dependency.version.sourceline if dependency.version else dependency.sourceline,
        is_transitive=False,
        is_dev=is_dev,
    )

    return JavaPackageSpec(
        composed_name=full_name,
        simple_name=name,
        version=str(version),
        location=new_location,
        ecosystem_data=JavaArchive(pom_properties=pom_properties),
    )


def _is_valid_version(version: str | None) -> bool:
    return version is not None and not version.startswith("${")



from collections.abc import Iterable

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection import toml
from labels.utils.strings import normalize_name


def parse_poetry_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = toml.parse_toml_with_tree_sitter(reader.read_closer.read())

    packages = _collect_packages(file_content, reader)
    relationships = _collect_relationships(file_content, packages)

    return packages, relationships


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    toml_pkgs = file_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []

    for raw_package in toml_pkgs:
        if not isinstance(raw_package, IndexedDict):
            continue
        name = str(raw_package.get("name", ""))
        version = str(raw_package.get("version", ""))

        if not version:
            continue

        new_location = (
            get_enriched_location(
                reader.location,
                line=raw_package.get_key_position("version").start.line,
            )
            if isinstance(raw_package, IndexedDict)
            else reader.location
        )

        package = new_python_package(name=name, version=version, location=new_location)
        if package:
            packages.append(package)

    return packages


def _get_dependencies(
    package: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(package, IndexedDict):
        return None

    package_name = package.get("name")
    if not isinstance(package_name, str):
        return None

    _pkg = _find_package_by_name(packages, package_name)

    deps = package.get("dependencies")
    if not isinstance(deps, IndexedDict):
        return None

    return _pkg, deps


def _find_package_by_name(packages: list[Package], name: str) -> Package | None:
    return next(
        (p for p in packages if p.name == normalize_name(name, PackageType.PythonPkg)), None
    )


def _collect_relationships(
    toml_content: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships = []
    toml_pkgs = toml_content.get("package")

    if not isinstance(toml_pkgs, IndexedList):
        return []

    for package in toml_pkgs:
        pkg_with_deps = _get_dependencies(package, packages)

        if not pkg_with_deps:
            continue

        pkg, deps = pkg_with_deps

        if not pkg or not deps:
            continue

        relationships.extend(_create_relationships_for_package(pkg, deps.keys(), packages))

    return relationships


def _create_relationships_for_package(
    pkg: Package, dependency_names: Iterable[str], packages: list[Package]
) -> list[Relationship]:
    relationships = []
    for dep_name in dependency_names:
        dep_pkg = _find_package_by_name(packages, dep_name)
        if dep_pkg:
            relationships.append(
                Relationship(
                    from_=dep_pkg.id_,
                    to_=pkg.id_,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
            )
    return relationships



import re
from typing import NamedTuple, NotRequired, TypedDict

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    add_dependency_to_root_relationship,
    build_relationships,
    create_root_package_from_package_json,
    find_root_location,
    get_dev_dependencies_from_package_json,
    get_direct_dependencies_from_package_json,
    group_packages_by_name,
)
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location


class YarnPackage(TypedDict):
    line: int
    version: str
    raw_key: NotRequired[str]  # Original key from yarn.lock header
    checksum: NotRequired[str]
    dependencies: NotRequired[list[tuple[str, str]]]
    integrity: NotRequired[str]
    resolution: NotRequired[str]
    resolved: NotRequired[str]


class PackageKey(NamedTuple):
    name: str
    version: str


class ParserState(NamedTuple):
    parsed_yarn_lock: dict[PackageKey, YarnPackage]
    current_package: str | None = None
    current_package_line: int | None = None
    current_package_version: str | None = None
    current_indentation: int | None = None
    current_key: str | None = None
    package_key: PackageKey | None = None
    current_raw_key: str | None = None  # Raw key from header line


def parse_yarn_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    yarn_lock_content = reader.read_closer.read()
    parsed_yarn_lock = _parse_yarn_file(yarn_lock_content)

    resolution_map = _build_resolution_map_from_parsed(parsed_yarn_lock)

    root_package = create_root_package_from_package_json(reader.location)
    has_package_json = root_package is not None

    direct_dependencies = get_direct_dependencies_from_package_json(reader.location)
    dev_dependencies = get_dev_dependencies_from_package_json(reader.location)

    packages = _extract_packages(
        parsed_yarn_lock,
        reader.location,
        direct_dependencies,
        dev_dependencies,
        has_package_json=has_package_json,
    )

    if root_package:
        packages.append(root_package)

    dependency_map = _build_dependency_map_yarn(parsed_yarn_lock, resolution_map)
    key_map = _build_key_map_yarn(parsed_yarn_lock)
    relationships = build_relationships(packages, dependency_map, reader.location, key_map)

    if root_package:
        root_relationships = _create_root_relationships(
            root_package, packages, direct_dependencies, reader.location
        )
        relationships.extend(root_relationships)

    return packages, relationships


def _parse_yarn_file(yarn_lock_content: str) -> dict[PackageKey, YarnPackage]:
    yarn_lock_lines = yarn_lock_content.strip().split("\n")
    initial_state = ParserState(parsed_yarn_lock={})

    final_state = _process_lines(yarn_lock_lines, initial_state)
    return final_state.parsed_yarn_lock


def _process_lines(lines: list[str], state: ParserState) -> ParserState:
    for index, line in enumerate(lines, 1):
        state = _process_line(line, index, state)
    return state


def _process_line(line: str, index: int, state: ParserState) -> ParserState:
    if not line:
        return state._replace(current_indentation=None)

    if line.startswith("#"):
        return state

    if not line.startswith(" "):
        return _handle_package_header(line, index, state)

    return _process_indented_line(line, state)


def _process_indented_line(line: str, state: ParserState) -> ParserState:
    if state.current_package and state.current_package_line and line.strip().startswith("version"):
        return _handle_version_line(line, state)

    if _is_start_of_list_line(state, line):
        return _handle_list_start(line, state)

    if _is_list_item_line(state, line):
        return _handle_list_item(line, state)

    if state.package_key:
        return _handle_property(line, state)

    return state


def _handle_package_header(line: str, index: int, state: ParserState) -> ParserState:
    current_package, current_package_line = _parse_current_package(line, index)
    raw_key = line.strip().rstrip(":").strip('"')
    return state._replace(
        current_package=current_package,
        current_package_line=current_package_line,
        current_package_version=None,
        package_key=None,
        current_raw_key=raw_key,
    )


def _handle_version_line(line: str, state: ParserState) -> ParserState:
    if not state.current_package or not state.current_package_line:  # pragma: no cover
        return state

    _, raw_version = _resolve_pair(line)
    version = raw_version.strip('"')
    package_key = PackageKey(name=state.current_package, version=version)

    new_package: YarnPackage = {
        "line": state.current_package_line,
        "version": version,
    }
    if state.current_raw_key:
        new_package["raw_key"] = state.current_raw_key

    new_parsed_lock = {**state.parsed_yarn_lock, package_key: new_package}

    return state._replace(
        parsed_yarn_lock=new_parsed_lock,
        current_package_version=version,
        package_key=package_key,
    )


def _handle_list_start(line: str, state: ParserState) -> ParserState:
    if not state.package_key:  # pragma: no cover
        return state

    indentation = _count_indentation(line)
    key = line.strip().split(":")[0]

    if key not in ("dependencies", "optionalDependencies"):
        return state._replace(current_indentation=None, current_key=None)

    existing_deps = state.parsed_yarn_lock[state.package_key].get("dependencies", [])

    updated_package: YarnPackage = {
        **state.parsed_yarn_lock[state.package_key],
        "dependencies": existing_deps if key == "optionalDependencies" else [],
    }
    new_parsed_lock = {**state.parsed_yarn_lock, state.package_key: updated_package}

    return state._replace(
        parsed_yarn_lock=new_parsed_lock,
        current_indentation=indentation,
        current_key=key,
    )


def _handle_list_item(line: str, state: ParserState) -> ParserState:
    if not state.package_key:  # pragma: no cover
        return state

    current_deps = state.parsed_yarn_lock[state.package_key].get("dependencies", [])
    new_dep = _resolve_pair(line)
    new_deps = [*current_deps, new_dep]

    updated_package: YarnPackage = {
        **state.parsed_yarn_lock[state.package_key],
        "dependencies": new_deps,
    }
    new_parsed_lock = {**state.parsed_yarn_lock, state.package_key: updated_package}

    return state._replace(parsed_yarn_lock=new_parsed_lock)


def _handle_property(line: str, state: ParserState) -> ParserState:
    if not state.package_key:  # pragma: no cover
        return state

    key, value = _resolve_pair(line)
    if key not in ("checksum", "integrity", "resolution", "resolved"):
        return state._replace(current_indentation=None)

    existing_package = state.parsed_yarn_lock[state.package_key]
    updated_package: YarnPackage = {**existing_package, key: value.strip('"')}  # type: ignore[misc]
    new_parsed_lock = {**state.parsed_yarn_lock, state.package_key: updated_package}

    return state._replace(parsed_yarn_lock=new_parsed_lock, current_indentation=None)


def _create_root_relationships(
    root_package: Package,
    packages: list[Package],
    direct_dependencies: set[str],
    location: Location,
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not (location.coordinates and location.coordinates.real_path):
        return relationships

    lockfile_path = location.coordinates.real_path
    root_location = find_root_location(root_package, lockfile_path)

    if not root_location:
        return relationships

    packages_by_name = group_packages_by_name(packages, root_package.id_)

    for dep_name in direct_dependencies:
        for dep_pkg in packages_by_name.get(dep_name, []):
            add_dependency_to_root_relationship(
                dep_pkg, root_package, root_location, lockfile_path, relationships
            )

    return relationships


def _extract_packages(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
    location: Location,
    direct_dependencies: set[str],
    dev_dependencies: set[str],
    *,
    has_package_json: bool,
) -> list[Package]:
    packages = []

    for pkg_info, item in parsed_yarn_lock.items():
        name = _get_name(pkg_info, item)
        version = item.get("version")

        if has_package_json:
            is_direct = name in direct_dependencies
            is_transitive: bool | None = not is_direct

            if is_direct:
                is_dev: bool | None = name in dev_dependencies
            else:
                is_dev = None
        else:
            is_transitive = None
            is_dev = None

        new_location = get_enriched_location(
            location,
            line=item["line"],
            is_transitive=is_transitive,
            is_dev=is_dev,
        )

        package = new_simple_npm_package(new_location, name, version)
        if package:
            packages.append(package)

    return packages


def _extract_patch_entry(entry: str) -> tuple[str, str] | None:
    if match := re.search(r"@patch:(.+?)@npm%3A([^#]+)", entry):
        pkg_name = match.group(1)
        version_range_encoded = match.group(2)
        return pkg_name, f"npm:{version_range_encoded}"
    return None


def _extract_npm_entry(entry: str) -> tuple[str, str] | None:
    parts = entry.split("@npm:")
    if len(parts) >= 2:
        return parts[0], f"npm:{parts[-1]}"
    return None


def _extract_scoped_entry(entry: str) -> tuple[str, str] | None:
    last_at = entry.rfind("@")
    pkg_name = entry[:last_at]
    version_range = entry[last_at + 1 :].rstrip(":")
    return pkg_name, version_range


def _extract_simple_entry(entry: str) -> tuple[str, str] | None:
    parts = entry.split("@", 1)
    if len(parts) == 2:
        return parts[0], parts[1].rstrip(":")
    return None


def _process_entry(entry: str, version: str, resolution_map: dict[tuple[str, str], str]) -> None:
    result = None

    if "@patch:" in entry:
        result = _extract_patch_entry(entry)
    elif "@npm:" in entry:
        result = _extract_npm_entry(entry)
    elif entry.startswith("@") and entry.count("@") >= 2:
        result = _extract_scoped_entry(entry)
    elif "@" in entry:
        result = _extract_simple_entry(entry)

    if result:
        pkg_name, version_range = result
        resolution_map[(pkg_name, version_range)] = version


def _build_resolution_map_from_parsed(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
) -> dict[tuple[str, str], str]:
    resolution_map: dict[tuple[str, str], str] = {}

    for item in parsed_yarn_lock.values():
        raw_key = item.get("raw_key")
        version = item.get("version")

        if not raw_key or not version:
            continue

        entries = raw_key.split(", ")
        for raw_entry in entries:
            entry = raw_entry.strip('"')
            _process_entry(entry, version, resolution_map)

    return resolution_map


def _build_dependency_map_yarn(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
    resolution_map: dict[tuple[str, str], str],
) -> DependencyMap:
    dependency_map: DependencyMap = {}

    for pkg_key, item in parsed_yarn_lock.items():
        pkg_name = _get_name(pkg_key, item)
        version = item.get("version")

        if "dependencies" in item and version is not None:
            deps_dict: dict[str, str] = {}

            for dep_name, dep_version in item["dependencies"]:
                clean_name = dep_name.strip('"')
                version_range = dep_version.strip('"')

                resolved_version = resolution_map.get((clean_name, version_range))

                deps_dict[clean_name] = resolved_version or version_range

            package_key = f"{pkg_name}@{version}"
            dependency_map[package_key] = deps_dict

    return dependency_map


def _build_key_map_yarn(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
) -> dict[tuple[str, int], str]:
    key_map: dict[tuple[str, int], str] = {}

    for pkg_key, item in parsed_yarn_lock.items():
        pkg_name = _get_name(pkg_key, item)
        version = item.get("version")
        line = item.get("line")

        if line is not None and version is not None:
            key_map[(pkg_name, line)] = f"{pkg_name}@{version}"

    return key_map


def _get_name(pkg_info: PackageKey, item: YarnPackage) -> str:
    if resolution := item.get("resolution"):
        is_scoped_package = resolution.startswith("@")
        if is_scoped_package:
            return f"@{resolution.split('@')[1]}"
        return resolution.split("@")[0]

    return pkg_info.name


def _parse_current_package(line: str, index: int) -> tuple[str | None, int | None]:
    line = line.strip()
    if match_ := re.match(r'^"?((?:@\w[\w\-\.]*/)?\w[\w\-\.]*)@', line):
        current_package = match_.groups()[0]
        current_package_line = index
    else:
        current_package = None
        current_package_line = None

    return current_package, current_package_line


def _resolve_pair(line: str) -> tuple[str, str]:
    line = line.strip()
    if ": " in line:
        key, value = line.split(": ")
        return key.strip(), value.strip()

    key, value = line.split(" ", maxsplit=1)
    return key.strip(), value.strip()


def _count_indentation(line: str) -> int:
    # Stripping the leading spaces and comparing the length difference
    return len(line) - len(line.lstrip(" "))


def _is_start_of_list_line(state: ParserState, line: str) -> bool:
    return bool(
        state.current_package and state.current_package_version and line.strip().endswith(":"),
    )


def _is_list_item_line(state: ParserState, line: str) -> bool:
    return bool(
        state.current_package
        and state.current_package_version
        and state.current_key
        and state.current_indentation
        and _count_indentation(line) > state.current_indentation,
    )



import re
from typing import NamedTuple

from gemfileparser import GemfileParser

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.ruby.package_builder import new_gemspec_package
from labels.parsers.cataloger.ruby.utils import format_ruby_version_constraints
from labels.parsers.cataloger.utils import get_enriched_location


class GemDependency(NamedTuple):
    name: str
    version_constraints: list[str]
    is_development: bool
    line: int


DEPENDENCY_PATTERN = re.compile(
    r"\w+\.add(_runtime|_development)?_dependency\s*\(?\s*"
    r"[\"']([^\"']+)[\"']"
    r"((?:\s*,\s*[\"'][^\"']+[\"'])*)"
    r"\s*\)?"
)

VERSION_CONSTRAINT_PATTERN = re.compile(r"[\"']([^\"']+)[\"']")


def parse_gemspec(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    dependencies = _extract_dependencies_by_line(content)
    packages = _build_packages_from_dependencies(dependencies, reader.location)

    return packages, []


def _extract_dependencies_by_line(content: str) -> list[GemDependency]:
    dependencies: list[GemDependency] = []

    for line_number, line in enumerate(content.splitlines(), start=1):
        preprocessed = GemfileParser.preprocess(line)
        if not preprocessed:
            continue

        dependency = _parse_dependency_line(preprocessed, line_number)
        if dependency:
            dependencies.append(dependency)

    return dependencies


def _parse_dependency_line(line: str, line_number: int) -> GemDependency | None:
    match = DEPENDENCY_PATTERN.search(line)
    if not match:
        return None

    dep_type = match.group(1) or ""
    is_development = dep_type == "_development"
    name = match.group(2)
    version_part = match.group(3)

    version_constraints = VERSION_CONSTRAINT_PATTERN.findall(version_part)
    if not version_constraints:
        return None

    return GemDependency(name, version_constraints, is_development, line_number)


def _build_packages_from_dependencies(
    dependencies: list[GemDependency], base_location: Location
) -> list[Package]:
    packages: list[Package] = []

    for dep in dependencies:
        version = format_ruby_version_constraints(dep.version_constraints)
        if not version:
            continue

        location = get_enriched_location(
            base_location,
            line=dep.line,
            is_transitive=False,
            is_dev=dep.is_development,
        )

        package = new_gemspec_package(name=dep.name, version=version, location=location)
        if package:
            packages.append(package)

    return packages



from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.ecosystem_data.debian import DpkgDBEntry
from labels.model.file import Location
from labels.model.package import Language, Package, PackageType
from labels.model.release import Release
from labels.parsers.cataloger.utils import (
    get_enriched_location,
    log_malformed_package_warning,
    purl_qualifiers,
)


def new_dpkg_package(
    entry: DpkgDBEntry,
    db_location: Location,
    release: Release | None = None,
) -> Package | None:
    name = entry.package
    version = entry.version

    if not name or not version:
        return None

    new_location = get_enriched_location(db_location)
    p_url = get_debian_package_url(entry, release)

    try:
        package = Package(
            name=name,
            version=version,
            p_url=p_url,
            locations=[new_location],
            type=PackageType.DebPkg,
            ecosystem_data=entry,
            found_by=None,
            language=Language.UNKNOWN_LANGUAGE,
        )
    except ValidationError as ex:
        log_malformed_package_warning(new_location, ex)
        return None

    return package


def get_debian_package_url(pkg: DpkgDBEntry, distro: Release | None = None) -> str:
    qualifiers = {"arch": pkg.architecture}
    if distro and (distro.id_ == "debian" or "debian" in (distro.id_like or [])):
        if distro.version_id:
            qualifiers["distro_version_id"] = distro.version_id
        qualifiers["distro_id"] = distro.id_
    if pkg.source:
        qualifiers["upstream"] = (
            f"{pkg.source}@{pkg.source_version}" if pkg.source_version else pkg.source
        )

    return PackageURL(  # type: ignore[misc]
        type="deb",
        namespace=distro.id_ if distro and distro.id_ else "",
        name=pkg.package,
        version=pkg.version,
        qualifiers=purl_qualifiers(qualifiers, distro),
    ).to_string()



from pydantic import ValidationError

from labels.model.package import Package, PackageType
from labels.model.syft_sbom import SyftArtifact
from labels.parsers.builder.utils import get_language, get_locations
from labels.parsers.cataloger.utils import log_malformed_package_warning


def builder(
    artifact: SyftArtifact,
    package_type: PackageType,
) -> Package | None:
    pkg_language = get_language(artifact.language)
    pkg_locations = get_locations(artifact.locations)
    pkg_name = artifact.name.lower()

    try:
        package = Package(
            name=pkg_name,
            version=artifact.version,
            p_url=artifact.p_url,
            locations=pkg_locations,
            type=package_type,
            found_by=artifact.found_by,
            language=pkg_language,
            syft_id=artifact.id_,
        )
    except ValidationError as ex:  # pragma: no cover
        log_malformed_package_warning(pkg_locations[0], ex)
        return None

    return package



import re

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    add_dependency_to_root_relationship,
    build_relationships,
    create_root_package_from_package_json,
    find_root_location,
    get_dev_dependencies_from_package_json,
    get_direct_dependencies_from_package_json,
    group_packages_by_name,
)
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter

VERSION_PATTERN = re.compile(r"^(.+)@(\d+\.\d+\.\d+(?:-[\w.]+)?)$")


def parse_bun_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    file_content = parse_json_with_tree_sitter(_strip_jsonc_comments(content))

    if not isinstance(file_content, IndexedDict):
        return [], []

    direct_dependencies = get_direct_dependencies_from_package_json(reader.location)
    dev_dependencies = get_dev_dependencies_from_package_json(reader.location)
    root_package = create_root_package_from_package_json(reader.location)

    packages = _collect_packages(
        reader.location,
        file_content,
        direct_dependencies,
        dev_dependencies,
    )

    if root_package:
        packages.append(root_package)

    dependency_map = _build_dependency_map(file_content)
    key_map = _build_key_map(file_content)

    relationships = build_relationships(packages, dependency_map, reader.location, key_map)

    if root_package:
        root_relationships = _create_root_relationships(
            root_package, packages, direct_dependencies, reader.location
        )
        relationships.extend(root_relationships)

    return packages, relationships


def _strip_jsonc_comments(content: str) -> str:
    lines = content.split("\n")
    result = []
    for line in lines:
        stripped = line.lstrip()
        if stripped.startswith("//"):
            continue
        result.append(line)
    return "\n".join(result)


def _collect_packages(
    location: Location,
    file_content: IndexedDict[str, ParsedValue],
    direct_dependencies: set[str],
    dev_dependencies: set[str],
) -> list[Package]:
    packages: list[Package] = []

    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return packages

    for package_key, package_entry in packages_dict.items():
        if not isinstance(package_entry, IndexedList):
            continue

        name_version = _parse_package_entry(package_entry)
        if not name_version:
            continue

        name, version = name_version

        is_direct = name in direct_dependencies
        is_transitive = not is_direct
        is_dev = name in dev_dependencies if is_direct else None

        position = packages_dict.get_key_position(package_key)
        new_location = get_enriched_location(
            location,
            line=position.start.line,
            is_transitive=is_transitive,
            is_dev=is_dev,
        )

        package = new_simple_npm_package(new_location, name, version)
        if package:
            packages.append(package)

    return packages


def _parse_package_entry(entry: IndexedList[ParsedValue]) -> tuple[str, str] | None:
    if len(entry) < 1:
        return None

    first_element = entry[0]
    if not isinstance(first_element, str):
        return None

    match = VERSION_PATTERN.match(first_element)
    if not match:
        return None

    return match.group(1), match.group(2)


def _build_dependency_map(file_content: IndexedDict[str, ParsedValue]) -> DependencyMap:
    dependency_map: DependencyMap = {}

    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return dependency_map

    for package_entry in packages_dict.values():
        result = _extract_package_dependencies(package_entry)
        if result:
            map_key, deps = result
            dependency_map[map_key] = deps

    return dependency_map


def _extract_package_dependencies(
    package_entry: ParsedValue,
) -> tuple[str, dict[str, str]] | None:
    if not isinstance(package_entry, IndexedList):
        return None

    name_version = _parse_package_entry(package_entry)
    if not name_version or len(package_entry) < 3:
        return None

    name, version = name_version
    deps_obj = package_entry[2]
    if not isinstance(deps_obj, IndexedDict):
        return None

    dependencies_dict = deps_obj.get("dependencies")
    if not isinstance(dependencies_dict, IndexedDict):
        return None

    deps = {
        dep_name: _extract_version(dep_version)
        for dep_name, dep_version in dependencies_dict.items()
        if isinstance(dep_version, str)
    }

    if not deps:
        return None

    return f"{name}@{version}", deps


def _build_key_map(
    file_content: IndexedDict[str, ParsedValue],
) -> dict[tuple[str, int], str]:
    key_map: dict[tuple[str, int], str] = {}

    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return key_map

    for package_key, package_entry in packages_dict.items():
        if not isinstance(package_entry, IndexedList):
            continue

        name_version = _parse_package_entry(package_entry)
        if not name_version:
            continue

        name, version = name_version
        position = packages_dict.get_key_position(package_key)
        key_map[(name, position.start.line)] = f"{name}@{version}"

    return key_map


def _extract_version(version_spec: str) -> str:
    match = re.match(r"^(\d+\.\d+\.\d+(?:-[\w.]+)?)", version_spec)
    if match:
        return match.group(1)
    return version_spec


def _create_root_relationships(
    root_package: Package,
    packages: list[Package],
    direct_dependencies: set[str],
    location: Location,
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not (location.coordinates and location.coordinates.real_path):
        return relationships

    lockfile_path = location.coordinates.real_path
    root_location = find_root_location(root_package, lockfile_path)

    if not root_location:
        return relationships

    packages_by_name = group_packages_by_name(packages, root_package.id_)

    for dep_name in direct_dependencies:
        for dep_pkg in packages_by_name.get(dep_name, []):
            add_dependency_to_root_relationship(
                dep_pkg, root_package, root_location, lockfile_path, relationships
            )

    return relationships



import tarfile
from contextlib import suppress
from io import TextIOWrapper

from labels.model.file import Coordinates, DependencyType, Location, LocationReadCloser, Scope

# IEC Sizes
BYTE = 1 << (0 * 10)
KIBYTE = 1 << (1 * 10)
MIBYTE = 1 << (2 * 10)
GIBYTE = 1 << (3 * 10)
TIBYTE = 1 << (4 * 10)
PIBYTE = 1 << (5 * 10)
EIBYTE = 1 << (6 * 10)

# SI Sizes
IBYTE = 1
KBYTE = IBYTE * 1000
MBYTE = KBYTE * 1000
GBYTE = MBYTE * 1000
TBYTE = GBYTE * 1000
PBYTE = TBYTE * 1000
EBYTE = PBYTE * 1000

bytes_size_table = {
    "b": BYTE,
    "kib": KIBYTE,
    "kb": KBYTE,
    "mib": MIBYTE,
    "mb": MBYTE,
    "gib": GIBYTE,
    "gb": GBYTE,
    "tib": TIBYTE,
    "tb": TBYTE,
    "pib": PIBYTE,
    "pb": PBYTE,
    "eib": EIBYTE,
    "eb": EBYTE,
    # Without suffix
    "": BYTE,
    "ki": KIBYTE,
    "k": KBYTE,
    "mi": MIBYTE,
    "m": MBYTE,
    "gi": GIBYTE,
    "g": GBYTE,
    "ti": TIBYTE,
    "t": TBYTE,
    "pi": PIBYTE,
    "p": PBYTE,
    "ei": EIBYTE,
    "e": EBYTE,
}


def extract_tar_file(tar_path: str, out_path: str) -> None:
    with tarfile.open(tar_path, "r") as tar:
        for member in tar.getmembers():
            member.name = member.name.lstrip("/")
            with suppress(Exception):
                tar.extract(member, path=out_path)


def new_location_from_image(
    access_path: str | None,
    layer_id: str,
    real_path: str | None = None,
) -> Location:
    if access_path and not access_path.startswith("/"):
        access_path = f"/{access_path}"
    return Location(
        coordinates=Coordinates(real_path=real_path or "", file_system_id=layer_id),
        access_path=access_path,
    )


def new_location(real_path: str) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
        ),
        access_path=real_path,
    )


def new_default_location_from_path(real_path: str, line: int | None = None) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
            line=line,
        ),
        access_path=real_path,
        scope=Scope.UNDETERMINABLE,
        dependency_type=DependencyType.UNDETERMINABLE,
        reachable_cves=[],
    )


def new_default_location_read_closer(
    str_path: str, read_closer: TextIOWrapper
) -> LocationReadCloser:
    location = new_default_location_from_path(str_path)
    return LocationReadCloser(location=location, read_closer=read_closer)



from typing import Any

import phpserialize

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.php.package_builder import new_package_from_pecl
from labels.parsers.cataloger.utils import get_enriched_location


def parse_pecl_serialized(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    unserialized_data = phpserialize.loads(reader.read_closer.read().encode(), decode_strings=True)

    packages = _collect_packages(unserialized_data, reader)

    return packages, []


def _collect_packages(unserialized_data: object, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []

    parsed_data = _php_to_python(unserialized_data)
    name = str(parsed_data.get("name", "")) or None
    version = str(parsed_data.get("version", {}).get("release", "")) or None

    new_location = get_enriched_location(reader.location)

    package = new_package_from_pecl(name=name, version=version, location=new_location)
    if package:
        packages.append(package)

    return packages


def _php_to_python(obj: Any) -> Any:  # type: ignore[explicit-any]  # noqa: ANN401
    if isinstance(obj, dict):
        return {_php_to_python(k): _php_to_python(v) for k, v in obj.items()}
    return obj



import re
from collections.abc import Iterator
from typing import Final, NamedTuple

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.golang.package_builder import new_go_package
from labels.parsers.cataloger.utils import get_enriched_location

GO_DIRECTIVE: re.Pattern[str] = re.compile(
    r"(?P<directive>require|replace) \(",
)
GO_MOD_DEP: re.Pattern[str] = re.compile(
    r"^\s+(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_REPLACE: re.Pattern[str] = re.compile(
    r"^\s+(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REP_DEP: re.Pattern[str] = re.compile(
    r"replace\s(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REQ_MOD_DEP: re.Pattern[str] = re.compile(
    r"require\s(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_VERSION: re.Pattern[str] = re.compile(
    r"\ngo (?P<major>\d)\.(?P<minor>\d+)(\.\d+)?\n",
)

MINIMUM_SUPPORTED_GO_VERSION: Final[tuple[int, int]] = (1, 17)


class ParseState(NamedTuple):
    replace_list: list[tuple[re.Match[str], int]]
    req_dict: dict[str, Package]


def parse_go_mod(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    file_content = reader.read_closer.read()
    if _is_go_version_supported(file_content):
        packages = list(_collect_packages(file_content, reader.location))

    return packages, []


def _handle_directive_block_line(
    line: str, line_number: int, directive: str, state: ParseState, location: Location
) -> str:
    if directive == "replace":
        if replace := GO_REPLACE.search(line):
            state.replace_list.append((replace, line_number))
            return directive
        return ""
    if matched := GO_MOD_DEP.search(line):
        _add_required_package(matched, state.req_dict, line_number, location)
        return directive
    return ""


def _collect_packages(
    file_content: str,
    location: Location,
) -> Iterator[Package]:
    directive: str = ""
    state = ParseState(replace_list=[], req_dict={})

    for line_number, line in enumerate(file_content.splitlines(), 1):
        if matched := GO_REQ_MOD_DEP.search(line):
            _add_required_package(matched, state.req_dict, line_number, location)
            continue
        if replace := GO_REP_DEP.search(line):
            state.replace_list.append((replace, line_number))
            continue
        if not directive:
            if match := GO_DIRECTIVE.match(line):
                directive = match.group("directive")
            continue
        directive = _handle_directive_block_line(line, line_number, directive, state, location)
    return _apply_replace_directives(state.req_dict, state.replace_list, location)


def _add_required_package(
    matched: re.Match[str],
    req_dict: dict[str, Package],
    line_number: int,
    parent_location: Location,
) -> None:
    product: str = matched.group("product")
    version: str = matched.group("version")

    new_location = get_enriched_location(parent_location, line=line_number, is_transitive=False)

    package = new_go_package(product, version, new_location)
    if package:  # pragma: no branch
        req_dict[product] = package


def _apply_replace_directives(
    req_dict: dict[str, Package],
    replace_list: list[tuple[re.Match[str], int]],
    parent_location: Location,
) -> Iterator[Package]:
    for matched, line_number in replace_list:
        match_dict = matched.groupdict()
        old_pkg, old_version = match_dict["old_prod"], match_dict["old_ver"]
        repl_pkg, version = match_dict["new_prod"], match_dict["new_ver"]

        if old_pkg not in req_dict:
            continue

        if old_version and not version:
            version = req_dict[old_pkg].version

        if not version or (old_version and req_dict[old_pkg].version != old_version):
            continue

        new_location = get_enriched_location(parent_location, line=line_number, is_transitive=False)

        package = new_go_package(repl_pkg, version, new_location)
        if package:  # pragma: no branch
            req_dict[old_pkg] = package

    return iter(req_dict.values())


def _is_go_version_supported(content: str) -> bool:
    go_version = GO_VERSION.search(content)
    if not go_version:
        return False

    major = int(go_version.group("major"))
    minor = int(go_version.group("minor"))

    return (major, minor) >= MINIMUM_SUPPORTED_GO_VERSION
